#include "handle_manager.h"

#define MAX_HANDLE          (65536 * 4)

int handle_mgr_create(handle_mgr_t *handle_mgr)
{
    long    *table = NULL;

    table = zx_calloc(MAX_HANDLE * sizeof(long));

    if(table == NULL)
    {
        return E_OUTOFMEMORY;
    }

    handle_mgr->max_handle  = MAX_HANDLE;
    handle_mgr->table       = table;
    handle_mgr->left_num    = handle_mgr->max_handle - 1;
    handle_mgr->free_start  = 1;

    handle_mgr->bitmap_lock = zx_create_spinlock(0);
    handle_mgr->bitmap_size = handle_mgr->max_handle >> 3;
    handle_mgr->bitmap      = zx_calloc(handle_mgr->bitmap_size);

    zx_set_bit(0, handle_mgr->bitmap);     // 0 will not used

    return S_OK;
}

void handle_mgr_destroy(handle_mgr_t *handle_mgr)
{
    if(handle_mgr->table)
    {       
        zx_free(handle_mgr->table);

        handle_mgr->max_handle = 0;
        handle_mgr->table      = NULL;
    }

    zx_free(handle_mgr->bitmap);

    zx_destroy_spinlock(handle_mgr->bitmap_lock);
}

unsigned int add_handle(handle_mgr_t *handle_mgr, unsigned int type, void *data)
{
    int             index     = 0;
    unsigned int    new_handle = 0;
    long            *table;
    unsigned long   flags;

    flags = zx_spin_lock_irqsave(handle_mgr->bitmap_lock);

    table = handle_mgr->table;

    index = zx_find_next_zero_bit(handle_mgr->bitmap, handle_mgr->max_handle, handle_mgr->free_start);

    if(index < handle_mgr->max_handle)
    {
        zx_set_bit(index, handle_mgr->bitmap);

        handle_mgr->left_num--;
        handle_mgr->free_start = (index + 1) % handle_mgr->max_handle;
        table[index] = (long)data;
        new_handle = index | (type << 24);
    }
    else
    {
        zx_info("no enough handle, left_num %d max_handle:%d.\n", handle_mgr->left_num, handle_mgr->max_handle);
        zx_assert(0, "no enough handle"); 
    }

    zx_spin_unlock_irqrestore(handle_mgr->bitmap_lock, flags);

    return new_handle;
}

void *get_from_handle_and_validate(handle_mgr_t *handle_mgr, unsigned int handle, unsigned int type)
{
    void *ptr = NULL;

    if((handle & 0xFF000000) == (type << 24))
    {
        ptr = get_from_handle(handle_mgr, handle);
    }

    return ptr;
}

void remove_handle(handle_mgr_t *handle_mgr, unsigned int handle)
{
    long            *table = NULL;
    unsigned int    index  = handle & 0x00FFFFFF;
    unsigned long   flags;

    flags = zx_spin_lock_irqsave(handle_mgr->bitmap_lock);

    table = handle_mgr->table;

    if(index < handle_mgr->max_handle)
    {
        handle_mgr->left_num++;

        zx_clear_bit(index, handle_mgr->bitmap);
        table[index] = 0;
    }
    else
    {
        zx_info("unknow handle: handle: %x, max_handle:%d.\n", handle, handle_mgr->max_handle);
        zx_assert(0, "unknow handle"); 
    }

    zx_spin_unlock_irqrestore(handle_mgr->bitmap_lock, flags);
}

